{*******************************************************}
{                                                       }
{       RichView                                        }
{       Demo: Dual unicode editor with synchronized     }
{       scrolling and highlighting.                     }
{                                                       }
{       Copyright (c) Sergey Tkachenko                  }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{                                                       }
{*******************************************************}

{ Limitation: this demo does not work properly with tables.
  Compile it with RVDONOTUSETABLES compiler define (see RV_Defs.inc).
  Note 1: when running in Delphi, an exception is possible in RVStyle1DrawParaBack.
    This exception is handled in the component and it's not dangerous.
  Note 2: possible optimization: in RichViewEditCaretMove, redraw an unfocused
    editor only if its HltItemNo is changed.
}

unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, RVStyle, RVScroll, RichView, RVEdit, CRVData;

type
  TForm1 = class(TForm)
    RichViewEdit1: TRichViewEdit;
    RichViewEdit2: TRichViewEdit;
    RVStyle1: TRVStyle;
    Splitter1: TSplitter;
    procedure RichViewEdit1VScrolled(Sender: TObject);
    procedure RichViewEdit2VScrolled(Sender: TObject);
    procedure RVStyle1DrawParaBack(Sender: TRVStyle; Canvas: TCanvas;
      ParaNo: Integer; ARect: TRect; var DoDefault: Boolean);
    procedure RichViewEditCaretMove(Sender: TObject);
    procedure RichViewEditEnter(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    { Private declarations }
    NoUpdate: Boolean;
    FocusedEditor: TObject;
    HltItemNo1, HltItemNo2: Integer;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

{============================ Scroll synchronization ==========================}

{ Returns the index of the first item of the same paragraph as ItemNo }
function GetFirstParagraphItemNo(rve: TCustomRichViewEdit; ItemNo: Integer): Integer;
begin
  while not rve.IsFromNewLine(ItemNo) do
    dec(ItemNo);
  Result := ItemNo;
end;

{ Returns the index of the last item of the same paragraph as ItemNo }
function GetLastParagraphItemNo(rve: TCustomRichViewEdit; ItemNo: Integer): Integer;
begin
  while (ItemNo+1<rve.ItemCount-1) and not rve.IsFromNewLine(ItemNo+1) do
    inc(ItemNo);
  Result := ItemNo;
end;

{ Returns the index of paragraph (more exactly - of paragraph section }
function GetParagraphNo(rve: TCustomRichViewEdit; ItemNo: Integer): Integer;
var i: Integer;
begin
  Result := 0;
  for i := 0 to ItemNo do
    if rve.IsFromNewLine(i) then
      inc(Result);
end;

{ Returns the index of the first item of the specified paragraph }  
function GetFirstItemNoInParagraph(rve: TCustomRichViewEdit; ParagraphNo: Integer): Integer;
var i: Integer;
begin
  Result := -1;
  for i := 0 to rve.ItemCount-1 do
    if rve.IsFromNewLine(i) then begin
      dec(ParagraphNo);
      if ParagraphNo=0 then begin
        Result := i;
        exit;
      end;
    end;
end;

{ Scrolls to the paragraph with index ParagraphNo, shifting the result by Delta }
procedure ScrollToParagraphNo(rve: TCustomRichViewEdit; ParagraphNo, Delta: Integer);
var ItemNo, x,y: Integer;
begin
  ItemNo := GetFirstItemNoInParagraph(rve, ParagraphNo);
  if ItemNo<0 then
    exit;
  rve.GetItemCoords(ItemNo, x, y);
  rve.ScrollTo(y+Delta);
end;

{ Scrolls rve2 to the same position as rve1 }
procedure SyncPosition(rve1, rve2: TCustomRichViewEdit);
var ItemNo, ParagraphNo, x, y: Integer;
begin
  ItemNo := rve1.FirstItemVisible;
  if ItemNo<0 then
    exit;
  ItemNo := GetFirstParagraphItemNo(rve1, ItemNo);
  rve1.GetItemClientCoords(ItemNo, x, y);
  ParagraphNo := GetParagraphNo(rve1, ItemNo);
  ScrollToParagraphNo(rve2, ParagraphNo, -y);
end;

{ Sync 2 by 1 }
procedure TForm1.RichViewEdit1VScrolled(Sender: TObject);
begin
  if NoUpdate then
    exit;
  NoUpdate := True;
  SyncPosition(RichViewEdit1, RichViewEdit2);
  NoUpdate := False;
end;

{ Sync 1 by 2 }
procedure TForm1.RichViewEdit2VScrolled(Sender: TObject);
begin
  if NoUpdate then
    exit;
  NoUpdate := True;
  SyncPosition(RichViewEdit2, RichViewEdit1);
  NoUpdate := False;
end;

{========================== Paragraph Highlight ===============================}
{ Loading default files }
procedure TForm1.FormCreate(Sender: TObject);
begin
  RichViewEdit1.LoadRVF(ExtractFilePath(Application.ExeName)+'en.rvf');
  RichViewEdit2.LoadRVF(ExtractFilePath(Application.ExeName)+'ru.rvf');
  RichViewEdit1.Format;
  RichViewEdit2.Format;
end;

{ Drawing paragraph background for both editors
  For focused editor, highlight the current paragraph. For another editor,
  highlights the paragraph containing item HltItemNo1/HltItemNo2. }
procedure TForm1.RVStyle1DrawParaBack(Sender: TRVStyle; Canvas: TCanvas;
  ParaNo: Integer; ARect: TRect; var DoDefault: Boolean);
var i, j: Integer;
    MakeRed: Boolean;
begin
  i := Sender.ItemNo;
  if Sender.RVData=RichViewEdit1.RVData then
    if FocusedEditor = RichViewEdit1 then
      j := RichViewEdit1.CurItemNo
    else
      j := HltItemNo1
  else
    if FocusedEditor = RichViewEdit2 then
      j := RichViewEdit2.CurItemNo
    else
      j := HltItemNo2;
  repeat
    MakeRed := i=j;
    if MakeRed then
      break;
    dec(i);
  until TCustomRVData(Sender.RVData).IsFromNewLine(i+1);
  if MakeRed then begin
    Canvas.Brush.Style := bsSolid;
    Canvas.Brush.Color := $CCCCFF;
    Canvas.FillRect(ARect);
  end;
end;
{ Updating HltItemNo for unfocused editor (rve2) when moving caret in
  the focused editor (rv1) }
procedure DoOnCaretMove(rve1, rve2: TCustomRichViewEdit;
  var HltItemNo: Integer);
var ItemNo, ParagraphNo: Integer;
begin
  ItemNo := rve1.CurItemNo;
  if ItemNo<0 then begin
    HltItemNo := -1;
    exit;
  end;
  ParagraphNo := GetParagraphNo(rve1, ItemNo);
  HltItemNo := GetFirstItemNoInParagraph(rve2, ParagraphNo);
  if HltItemNo>=0 then
    HltItemNo := GetLastParagraphItemNo(rve2, HltItemNo);
end;

{ When caret is moved... }
procedure TForm1.RichViewEditCaretMove(Sender: TObject);
begin
  if FocusedEditor = RichViewEdit2 then
    DoOnCaretMove(RichViewEdit2, RichViewEdit1, HltItemNo1)
  else
    DoOnCaretMove(RichViewEdit1, RichViewEdit2, HltItemNo2);
  RichViewEdit1.Invalidate;
  RichViewEdit2.Invalidate
end;

{ Updating focused editor reference }
procedure TForm1.RichViewEditEnter(Sender: TObject);
begin
  FocusedEditor := Sender;
end;


end.
